//<c>   Copyright 1980-2000 by Gerry J. Danen; all rights reserved.

//<t>   Create a File with C Function Prototypes


#include"gd_tools.h"
#include"gd_ansi.h"


#define pgVERS_ "V2.21"
#define pgTITL_ "Create C Function Prototypes"
#define pgCOPR_ "1993"


#if     (! LARGE_MODEL)
  #error LARGE MEMORY MODEL REQUIRED
#endif


#define ISCSYM(x)       ((x) > 0 && (isalnum(x) || (x) == '_'))
#define NOT_SO_GOOD     ( (word_t *) -1 )
#define MAXPARAM        99                      // max. number of parameters to a function (orig: 20)
#define NEWBUFSIZ       (20480*sizeof(char))    // new buffer size
#define MAXLINES        10000                   // maximum prototype definitions


char    szHdr[] = "//\tPrototypes for %s\n\n" ;
UCHAR   fl_in[FS_LEN], fl_in_lcase[FS_LEN], fl_out[FS_LEN], fl_out_def[] = "CDECL.$" ;
UCHAR   sFUNC[81], sFunc[81], sDecl[1001], sParlist[2001], sBuffer[3001] ;

char    *iobuf ;                        // I/O buffer
char    *strings[MAXLINES] ;            // Array of pointers to strings
int     index ;                         // strings[] array index

int     inquote  = 0;                   // in a quote??
int     newline_seen = 1;               // are we at the start of a line
int     glastc   = ' ';                 // last char. seen by getsym()
int     files_done = 0 ;                // files processed
int     longest_funcname = 0 ;          // length of longest function name
long    linenum  = 1L;                  // line number in current file

BOOL    dostatic = FALSE ;              // do static functions?
BOOL    dohead   = TRUE ;               // do file headers?
BOOL    xref_mode = FALSE ;
BOOL    show_fnames = FALSE ;

FILE    *fOut ;

typedef struct word_list_struct_
{
    struct word_list_struct_ *next;
    char   string[1];
}   word_t;


//      Prototypes
void    addword(word_t *w,char *s);
void    add_fdef_to_table(STRING fdef);
int     compare(const void *arg1,const void *arg2);
void    emit(word_t *wlist,word_t *plist,long startline);
int     fnextch(FILE *f);
int     foundin(word_t *w1,word_t *w2);
void    getdecl(FILE *f);
word_t * getparamlist(FILE *f);
int     getsym(char *buf,FILE *f);
int     length_of_list(word_t *w);
int     nextch(FILE *f);
int     ngetc(FILE *f);
void    process_file_spec(STRING fspec);
int     skipit(char *buf,FILE *f);
void    syntax(void);
void    typefixhack(word_t *w);
word_t * typelist(word_t *p);
word_t * word_alloc(char *s);
word_t * word_append(word_t *w1,word_t *w2);
void    word_free(word_t *w);


void    syntax( void )
{
    logon(1);
    printf(
        "\n%s generates a prototype include file, or a function cross-reference,\n"
        "depending on switch settings.\n\n"
        "The program does NOT process #if statements, so any mismatched {} caused by\n"
        "conditional # statements may screw up the resulting file.\n\n"
        "Wildcards are allowed, so you can process *.c, for example.\n\n"
        "Syntax: %s [/s][/n][/x][/fFile][files ...]\n\n"
        "\t/s: include declarations for static functions\n"
        "\t/n: no header comment in output file (for multi-file processing)\n"
        "\t/x: xref mode (includes filename and linenumbers)\n"
        "\t/fFile: output to File (default is `%s')\n\n",
        get_global_program_name(), get_global_program_name(), fl_out_def );
    exit( GD_ERROR_HELP );
}


//      Routines for manipulating lists of words

word_t  *word_alloc( char *s )
{
    word_t *w ;

    w = (word_t *) malloc(sizeof(word_t) + strlen(s) + 1);
    if ( w == NULPTR )
    {
        fprintf( stderr, "\nNot enough memory to add word \"%s\" to list!\a\n", s );
        exit( GD_ERROR_NO_MEMORY );
    }
    strcpy( w->string, s );
    w->next = NULPTR ;
    return( w );
}


void    word_free( word_t *w )
{
    word_t *oldw ;

    while ( w )
    {
        oldw = w;
        w = w->next;
        free(oldw);
    }
}


//      Return the length of a list; empty words are not counted

int     length_of_list( word_t *w )
{
    register int count = 0;

    while ( w )
    {
        if ( *w->string )
            count += 1 ;
        w = w->next;
    }
    return( count );
}


//      Append two lists, and return the result

word_t  *word_append( word_t *w1, word_t *w2 )
{
    word_t *r, *w ;

    r = w = word_alloc("");

    while ( w1 )
    {
        w->next = word_alloc(w1->string);
        w = w->next;
        w1 = w1->next;
    }

    while ( w2 )
    {
        w->next = word_alloc(w2->string);
        w = w->next;
        w2 = w2->next;
    }

    return( r );
}


//      See if the last entry in w2 is in w1

int     foundin( word_t *w1, word_t *w2 )
{
    while ( w2->next )
        w2 = w2->next ;

    while ( w1 )
    {
        if ( ! strcmp(w1->string, w2->string) )
            return( 1 );
        w1 = w1->next;
    }
    return( 0 );
}


//      Add the string s to the given list of words

void    addword( word_t *w, char *s )
{
    while ( w->next )
        w = w->next;
    w->next = word_alloc( s );
}


//      Given a list representing a type and a variable name, extract just
//      the base type, e.g. "struct word_list_struct_ *x" would yield
//      "struct word_list_struct_"

word_t  *typelist( word_t *p )
{
    word_t *w, *r ;

    r = w = word_alloc( "" );
    while ( p && p->next )
    {
        if ( p->string[0] && (! ISCSYM(p->string[0])) )
            break;
        w->next = word_alloc( p->string );
        w = w->next ;
        p = p->next ;
    }
    return( r );
}


//      Promote formal parameters of type "char", "unsigned char", "short",
//      or "unsigned short" to "int".

void    typefixhack( word_t *w )
{
    word_t *oldw = NULPTR ;

    while ( w )
    {
        if ( *w->string )
        {
            if ( ((! strcmp(w->string, "char")) ||
                (! strcmp(w->string, "short")) )
                && (length_of_list(w->next) < 2) )
            {
                if ( oldw && !strcmp(oldw->string, "unsigned") )
                {
                    oldw->next = w->next;
                    free( w );
                    w = oldw;
                }
                strcpy( w->string, "int" );
            }
        }
        w = w->next;
    }
}


//      Read a character: if it's a newline, increment the line count

int     ngetc( FILE *f )
{
    int c;

    c = getc(f);
    if ( c == '\n' )
        linenum += 1 ;
    return( c );
}


//      Read the next character from the file.  If the character is '\' then
//      read and skip the next character.  Any comment sequence is converted to
//      a blank.

int     fnextch( FILE *f )
{
    int         c, lastc ;
    BOOL        incomment ;

    c = ngetc( f );
    while ( c == '\\' )
    {
        c = ngetc(f);                           // skip a character
        c = ngetc(f);
    }
    if ( c == '/' && ! inquote )
    {
        c = ngetc(f);
        if ( c == '*' )                         // comment type is "/* ... */"
        {
            incomment = TRUE ;
            c = ' ';
            while ( incomment )
            {
                lastc = c;
                c = ngetc(f);
                if ( lastc == '*' && c == '/' )
                    incomment = FALSE ;
                else
                if ( c < 0 )
                    return( c );
            }
            return( fnextch(f) );
        }
        else
        if ( c == '/' )                         // comment type is "//"
        {
            incomment = TRUE ;
            while ( incomment )
            {
                c = ngetc(f);
                if ( c == '\n' )
                {
                    incomment = FALSE ;
                }
                else
                if ( c < 0 )
                    return( c );
            }
            return( fnextch(f) );
        }
        else
        {
            if ( c == '\n' )
                linenum--;
            ungetc(c, f);
            return( '/' );
        }
    }
    return( c );
}


//      Get the next "interesting" character. Comments are skipped, and strings
//      are converted to "0". Also, if a line starts with "#" it is skipped.

int     nextch( FILE *f )
{
    int c;

    c = fnextch(f);
    if ( newline_seen && c == '#' )
    {
        do  {
            c = fnextch(f);
        }   while (c >= 0 && c != '\n');
        if ( c < 0 )
            return( c );
    }
    newline_seen = (c == '\n');

    if ( c == '\'' || c == '\"' )
    {
        inquote = c;
        while ( (c = fnextch(f)) >= 0 )
        {
            if ( c == inquote )
            {
                inquote = 0;
                return( '0' );
            }
        }
    }
    return( c );
}


//      Get the next symbol from the file, skipping blanks.
//      Return 0 if OK, -1 for EOF.
//      Also collapses everything between { and }

int     getsym( char *buf, FILE *f )
{
    register int c;
    int inbrack = 0;

    c = glastc;
    while ((c > 0) && isspace(c))
    {
        c = nextch(f);
    }
    if ( c < 0 )
    {
        return( -1 );
    }
    if ( c == '{' )
    {
        inbrack = 1;
        while (inbrack)
        {
            c = nextch(f);
            if (c < 0)
            {
                glastc = c;
                return( c );
            }
            if ( c == '{' )
                inbrack++;
            else
            if ( c == '}' )
                inbrack--;
        }
        strcpy( buf, "{}" );
        glastc = nextch(f);
        return( 0 );
    }
    if ( ! ISCSYM(c) )
    {
        *buf++ = c;
        *buf = 0;
        glastc = nextch(f);
        return( 0 );
    }
    while ( ISCSYM(c) )
    {
        *buf++ = c;
        c = nextch(f);
    }
    *buf = 0;
    glastc = c;
    return( 0 );
}


//      Skip until a ";" or the end of a function declaration is seen

int     skipit( char *buf, FILE *f )
{
    int i;

    do  {
        i = getsym(buf, f);
        if ( i < 0 )
            return( i );
    }   while (*buf != ';' && *buf != '{');
    return( 0 );
}


//      Get a parameter list; when this is called the next symbol in line
//      should be the first thing in the list.

word_t  *getparamlist( FILE *f )
{
    static word_t *pname[MAXPARAM];     // parameter names
    word_t      *tlist,                 // type name
                *plist;                 // temporary
    int         np = 0;                 // number of parameters
    int         typed[MAXPARAM];        // parameter has been given a type
    int         tlistdone;              // finished finding the type name
    int         sawsomething;
    int         i;
    int         inparen = 0;
    char buf[80];

    for (i = 0; i < MAXPARAM; i++)
        typed[i] = 0;

    plist = word_alloc("");

    //  first, get the stuff inside brackets (if anything)

    sawsomething = 0;           // gets set nonzero when we see an arg
    for (;;)
    {
        if ( getsym(buf, f) < 0 )
            return( NULPTR );
        if (*buf == ')' && (--inparen < 0))
        {
            if ( sawsomething )         /* if we've seen an arg */
            {
                pname[np] = plist;
                plist = word_alloc("");
                np++;
            }
            break;
        }
        if ( *buf == ';' )      /* something weird */
        {
            return( NOT_SO_GOOD );
        }
        sawsomething = 1;       /* there's something in the arg. list */
        if ( *buf == ',' && inparen == 0 )
        {
            pname[np] = plist;
            plist = word_alloc("");
            np++;
        }
        else
        {
            addword(plist, buf);
            if (*buf == '(')
                inparen++;
        }
    }

    //  next, get the declarations after the function header

    inparen = 0;

    tlist = word_alloc("");
    plist = word_alloc("");
    tlistdone = 0;
    sawsomething = 0;
    for ( ;; )
    {
        if ( getsym(buf, f) < 0 )
            return( NULPTR );

        /* handle a list like "int x,y,z" */
        if (*buf == ',' && !inparen)
        {
            if ( ! sawsomething )
                return( NULPTR );
            for (i = 0; i < np; i++)
            {
                if (!typed[i] && foundin(plist, pname[i]))
                {
                    typed[i] = 1;
                    word_free(pname[i]);
                    pname[i] = word_append(tlist, plist);
                    /* promote types */
                    typefixhack(pname[i]);
                    break;
                }
            }
            if (!tlistdone)
            {
                tlist = typelist(plist);
                tlistdone = 1;
            }
            word_free(plist);
            plist = word_alloc("");
        }
        /* handle the end of a list */
        else
        if (*buf == ';')
        {
            if ( ! sawsomething )
                return( NOT_SO_GOOD );
            for ( i = 0; i < np; i++ )
            {
                if (!typed[i] && foundin(plist, pname[i]))
                {
                    typed[i] = 1;
                    word_free(pname[i]);
                    pname[i] = word_append(tlist, plist);
                    typefixhack(pname[i]);
                    break;
                }
            }
            tlistdone = 0;
            word_free(tlist);
            word_free(plist);
            tlist = word_alloc("");
            plist = word_alloc("");
        }
        //  handle the beginning of the function
        else
        if (!strcmp(buf, "{}"))
            break;
        //  otherwise, throw the word into the list (except for "register")
        else
        if (strcmp(buf, "register"))
        {
            sawsomething = 1;
            addword(plist, buf);
            if (*buf == '(')
                inparen++;
            if (*buf == ')')
                inparen--;
        }
    }

    /* Now take the info we have and build a prototype list */

    /* empty parameter list means "void" */
    if (np == 0)
        return( word_alloc("void") );

    plist = tlist = word_alloc("");
    for (i = 0; i < np; i++)
    {

        /* If no type provided, make it an "int" */
        if ( !(pname[i]->next) ||
            (!(pname[i]->next->next)&&strcmp(pname[i]->next->string, "void")))
        {
            addword(tlist, "int");
        }
        while (tlist->next) tlist = tlist->next;
        tlist->next = pname[i];
        if (i < np - 1)
            addword( tlist, "," );
    }
    return( plist );
}


void    getdecl( FILE *f )              // get all the function declarations
{
    word_t *plist, *wlist = NULL ;
    char buf[80];
    int sawsomething;
    int oktoprint;
    long startline;         /* line where declaration started */

again:
    word_free(wlist);
    wlist = word_alloc("");
    sawsomething = 0;
    oktoprint = 1;

    for(;;)
    {
        if ( getsym(buf,f) < 0 )
        {
            return ;
        }
        //  try to guess when a declaration is not an external function definition
        if ((!strcmp(buf, ",")) || (!strcmp(buf, "{}")) ||
            (!strcmp(buf, "=")) || (!strcmp(buf, "typedef")) ||
            (!strcmp(buf, "extern")) )
        {
            skipit(buf, f);
            goto again;
        }
        if ( ! dostatic && ((!strcmp(buf, "static")) || (!strcmp(buf,"LOCALPROC"))) )
        {
            oktoprint = 0;
        }
        /* for the benefit of compilers that allow "inline" declarations */
        if (!strcmp(buf, "inline") && !sawsomething)
            continue;
        if ( ! strcmp(buf, ";") )
            goto again;

        /* A left parenthesis *might* indicate a function definition */
        if ( ! strcmp(buf, "(") )
        {
            startline = linenum;
            if ( (! sawsomething) || (! (plist = getparamlist(f))) )
            {
                skipit(buf, f);
                goto again ;
            }
            if ( plist == NOT_SO_GOOD )
                goto again ;

            if ( oktoprint )                    // looks like we got what we want
                emit( wlist, plist, startline );
            word_free(plist);
            goto again;
        }
        addword(wlist, buf);
        sawsomething = 1;
    }
}


//      Comparison function, called by qsort()

int     compare( const void *arg1, const void *arg2 )
{
    return( strcmp( *(char **)arg1, *(char **)arg2 ) );
}


//      Add function definition to the to-be-sorted table

void    add_fdef_to_table( STRING fdef )
{
    st_trim( fdef );

    if ( ! strlen(fdef) )
        return ;

    if ( index < MAXLINES )
    {
        if ( (strings[index++] = _strdup(fdef)) == NULL )
        {
            fprintf( stderr, "\nOut of memory: %s:%s\n", fl_in, fdef );
            return ;
        }
    }
}


void    fix_sDecl_spacing( void )               // make it prettier
{
    UCHAR       w1[11] ;

    st_reduce_spaces( sDecl );
    st_trim( sDecl );

    if ( xref_mode )                            // no need to do anything else
        return ;

    if ( st_len_word(sDecl) > 7 )
    {
        ST_CAT( sDecl, " " );
        return ;
    }
    st_get_word( w1, sDecl );
    st_remove_word( sDecl );
    if ( sDecl[0] == '*' )
    {
        st_remove_word( sDecl );
        strcat( w1, " *" );
    }
    if ( strlen(sDecl) )
        sprintf( sBuffer, "%-8s%s ", w1, sDecl );
    else
        sprintf( sBuffer, "%-7s ", w1 );
    ST_CPY( sDecl, sBuffer );
}


//      Emit a function declaration.  The attributes and name of the function
//      are in wlist; the parameters are in plist.

void    emit( word_t *wlist, word_t *plist, long startline )
{
    word_t *w ;
    char *p ;
    int count = -1 ;

    sFunc[0] = sDecl[0] = sParlist[0] = EOS ;

    for ( w = wlist; w; w = w->next )
    {
        count++ ;
        p = w->string;
        if ( st_eq(p, "#") )                    // don't do preprocessor stuff
            return ;
// printf( "{%s} ", p );
        strcat( sDecl, p );
        strcat( sDecl, " " );
    }

    if ( st_eq(p, "main") )                     // don't do main()
        return ;

    strcpy( sFunc, p );                         // save function name
    if ( ! show_fnames )
        printf( "%-23s\t", sFunc );

    if ( count == 1 )
    {
        sprintf( sParlist, "int %s", sDecl );
        strcpy( sDecl, sParlist );
        sParlist[0] = EOS ;
    }

    for ( w = plist; w; w = w->next )
    {
        p = w->string;
// printf( "[%s] ", p );
        if ( *p == ',' || *p == '[' || *p == '(' || *p == ')' )
            st_rtrim( sParlist );
        strcat( sParlist, p );
        if ( ISCSYM( *p ) )
            strcat( sParlist, " " );
    }
    st_trim( sFunc );
    st_trim( sDecl );
    st_trim( sParlist );
    sDecl[strlen(sDecl) - strlen(sFunc)] = EOS ;
    st_reduce_spaces( sParlist );
    fix_sDecl_spacing();
    strcpy( sFUNC, sFunc );
    strupr( sFUNC );

    if ( xref_mode )
    {
        sprintf( sBuffer, "%-32.32s%-32s%-12.12s%5ld %s %s(%s",
                 sFUNC, sFunc, fl_in_lcase, startline, sDecl, sFunc, sParlist );
    }
    else
    {
//      sprintf( sBuffer, "%-32.32s%s%s%s(%s",
//               sFUNC, sDecl, (strlen(sDecl) < 8 ? "\t" : " "), sFunc, sParlist );
        sprintf( sBuffer, "%-32.32s%s%s(%s",
                 sFUNC, sDecl, sFunc, sParlist );
    }

    if ( (int) strlen(sFunc) > longest_funcname )
        longest_funcname = strlen(sFunc);

    add_fdef_to_table( sBuffer );
}


void    process_file_spec( STRING fspec )
{
    struct find_t dta ;
    int         done ;
    FILE        *f ;

    if ( (-1 != st_ch_index('\\', fspec )) ||
         (-1 != st_ch_index( ':', fspec )) ||
         (-1 != st_ch_index( '*', fspec )) )
    {
        show_fnames = TRUE ;
        dohead = FALSE ;
    }

    done = _dos_findfirst( fspec, REGULAR_FILES, &dta );
    while ( ! done )
    {
        ST_CPY( fl_in, dta.name );
        strupr( fl_in );
        strcpy( fl_in_lcase, fl_in );
        strlwr( fl_in_lcase );

        if ( ! (f = fopen(fl_in, "r")) )
        {
            perror( fl_in );
            exit( GD_ERROR_CANNOT_OPEN_FOR_READ );
        }
        if ( iobuf )
            setvbuf( f, iobuf, _IOFBF, NEWBUFSIZ );
        if ( dohead )
            fprintf( fOut, szHdr, fl_in );
        if ( show_fnames )
            printf( "\r%-13s\r", fl_in_lcase );
        linenum = 1;
        newline_seen = 1;
        glastc = ' ';
        files_done += 1 ;
        getdecl(f);
        fclose(f);
        done = _dos_findnext( &dta );
    }
}


void    main( int argc, char *argv[] )
{
    char *t ;
    int i ;

    tt_init( SHAREWARE_VERSION, 0, pgVERS_, pgTITL_, pgCOPR_ );
    ansi_cls( 1, stdout );
    logon(0);

    strcpy( fl_out, fl_out_def );               // default output file name

    argv++;
    argc--;

    iobuf = malloc( NEWBUFSIZ );
    if ( iobuf == NULPTR )
    {
        fprintf( stderr, "\nNot enough memory for expanded I/O buffer!\a\n"
                         "Continuing with standard buffer.\n" );
    }

    while ( *argv && (**argv == '-' || **argv == '/') )
    {
        t = *argv++;
        --argc;
        t++;
        switch ( *t )
        {
            case 'n' :
                dohead = FALSE ;
                break ;
            case 's' :
                dostatic = TRUE ;
                break ;
            case 'x' :
                xref_mode = TRUE ;
                dohead = FALSE ;
                show_fnames = TRUE ;
                break ;
            case 'f' :
                t++ ;
                strcpy( fl_out, t );
                break ;
            default :
                syntax();
                break ;
        }
    }

    if ( argc == 0 )                            // no more args?
        syntax();

    fOut = fl_create_or_replace_tty( fOut, fl_out, "w+", YES, YES );

    while ( argc > 0 && *argv )                 // process all files
    {
        process_file_spec( *argv );
        argc--;
        argv++;
    }

    printf( "\rSorting..." );
    qsort( &strings, index, sizeof(strings[0]), compare );

    for ( i = 0; i < index; i++ )               // produce output file
    {
        strcpy( sBuffer, strings[i] );
        fprintf( fOut, "%s);\n", &sBuffer[32] );
    }
    fclose( fOut );

    printf( "\r%d File%s processed, containing %d function definition%s.\n",
            files_done, (files_done == 1 ? "" : "s"),
            index, (index == 1 ? "" : "s") );
    printf( "The longest function name is %d characters long.\n", longest_funcname );

    exit( GD_ERROR_NONE );
}
